import numpy as np
import matplotlib.pyplot as plt

from seaborn_extensions.types import DataFrame, Figure
from seaborn_extensions.utils import get_grid_dims, log_pvalues


def volcano_plot(stats: DataFrame, x="A", y="B", features=None) -> Figure:
    if features is None:
        stats["features"] = stats.index
    else:
        stats["features"] = stats[features]
    ### volcano plot
    combs = stats[["A", "B"]].drop_duplicates().reset_index(drop=True)
    stats["hedges"] *= -1
    stats["logp-unc"] = log_pvalues(stats["p-unc"].fillna(1))
    stats["logp-cor"] = log_pvalues(stats["p-cor"].fillna(1))
    stats["p-cor-plot"] = (stats["logp-cor"] / stats["logp-cor"].max()) * 5
    n, m = get_grid_dims(combs.shape[0])
    fig, axes = plt.subplots(n, m, figsize=(4 * m, 4 * n), squeeze=False)
    for idx, (a, b) in combs.iterrows():
        ax = axes.flatten()[idx]
        p = stats.query(f"{x} == '{a}' & {y} == '{b}'")
        v = p["hedges"].abs().max()
        ax.axvline(0, linestyle="--", color="grey")
        ax.scatter(
            p["hedges"],
            p["logp-unc"],
            c=p["hedges"],
            s=5 + (2 ** p["p-cor-plot"]),
            cmap="coolwarm",
            vmin=-v,
            vmax=v,
        )
        ax.set(title=f"{b} / {a}", ylabel=None, xlabel=None)
        for t in p.query(f"`p-cor` < 0.05").index:
            ax.text(
                p.loc[t, "hedges"],
                p.loc[t, "logp-unc"],
                s=p.loc[t, "features"],
                ha="right" if p.loc[t, "hedges"] > 0 else "left",
            )
    for ax in axes.flatten()[idx + 1 :]:
        ax.axis("off")

    for ax in axes[:, 0]:
        ax.set(ylabel="-log10(p-val)")
    for ax in axes[-1, :]:
        ax.set(xlabel="Hedges' g")
    return fig



"""
A type of plot that combines swarms and box(en)/bar plots in an overlaid fashion.
"""

import typing as tp
import itertools
import warnings

import numpy as np
import pandas as pd
import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns
import pingouin as pg

from seaborn_extensions.types import DataFrame, Axis, Figure, Iterables
from seaborn_extensions.utils import get_grid_dims, filter_kwargs_by_callable


"""
# Demo with various tests available in Pingouin:
data = pd.DataFrame(
    {"cont": np.random.random(20), "cat": np.random.choice(["a", "b"], 20)}
)
data.loc[data["cat"] == "b", "cont"] *= 5
x = 'cat'
y = 'cont'

fig, stats = swarmboxenplot(data=data, x='x', y='y', hue='h')

fig, stats = swarmboxenplot(data=data, x=x, y=y)

pg.ttest(*data.groupby(x)[y].apply(lambda x: list(x)))
pg.mwu(*data.groupby(x)[y].apply(lambda x: list(x)))
pg.kruskal(data=data, between=x, dv=y)
pg.pairwise_ttests(data=data, between=x, dv=y, parametric=True)  # same as T-test
pg.pairwise_ttests(data=data, between=x, dv=y, parametric=False)  # same as MWU

c = data[x].astype(pd.CategoricalDtype())
pg.linear_regression(pd.concat((c[col].cat.codes.rename(col) for col in c), axis=1), data[y])
pg.linear_regression(pd.get_dummies(c), data[y])
pg.logistic_regression(data[y], c.cat.codes)
"""


def swarmboxenplot(
    data: DataFrame,
    x: str,
    y: tp.Union[str, Iterables],
    hue: str = None,
    swarm: bool = True,
    boxen: bool = True,
    bar: bool = False,
    orient: str = "vertical",
    plot: bool = True,
    ax: tp.Union[Axis, tp.Sequence[Axis]] = None,
    test: tp.Union[bool, str] = "mann-whitney",
    to_test: str = "all",
    multiple_testing: tp.Union[bool, str] = "fdr_bh",
    test_upper_threshold: float = 0.05,
    test_lower_threshold: float = 0.01,
    plot_non_significant: bool = False,
    plot_kws: tp.Dict[str, tp.Any] = None,
    test_kws: tp.Dict[str, tp.Any] = None,
    fig_kws: tp.Dict[str, tp.Any] = None,
) -> tp.Optional[tp.Union[Figure, DataFrame, tp.Tuple[Figure, DataFrame]]]:
    """
    A categorical plot that overlays individual observations
    as a swarm plot and summary statistics about them in a boxen plot.

    In addition, this plot will test differences between observation
    groups and add lines representing a significant difference between
    them.

    Parameters
    ----------
    data: pd.DataFrame
        A dataframe with data where the rows are the observations and
        columns are the variables to group them by.
    x: str
        The categorical variable.
    y: str | list[str]
        The continuous variable to plot.
        If more than one is given, will ignore the `ax` attribute and
        return figure with a subplot per each `y` variable.
    hue: str, optional
        An optional categorical variable to further group observations by.
    swarm: bool
        Whether to plot individual observations as a swarmplot.
    boxen: bool
        Whether to plot summary statistics as a boxenplot.
    bar: bool
        Whether to plot summary statistics as a barplot.
    orient: str
        Whether the plot should be oriented horizontally or vertically with relation to the numeric values `y`.
            - 'vertical': y-axis is `y` variable (numeric).
            - 'horizontal': x-axis is `y` variable (numeric).
        Default is 'vertical'.
    ax: matplotlib.axes.Axes, optional
        An optional axes to draw in.
    test: bool | str
        Whether to test differences between observation groups.
        If `False`, will not return a dataframe as well.
        If a string is passed, will perform test accordingly. Available tests:
            - 't-test'
            - 'mann-whitney'
            - 'kruskal'
        Default is a parwise 'mann-whitney' test with p-value adjustment.
    to_test: str
        Whether to test all possible combinations or just within `hue` groups for each `x`.
        Only relevant when `hue` is not None.
            - 'all': a model "y ~ x * hue", i.e. test between `x` groups, and within `hue` for each `x`.
            - 'hue': a model "y ~ x | hue", i.e. test within `hue` for each `x`.
    multiple_testing: str
        Method for multiple testing correction.
    test_upper_threshold: float
        Upper theshold to consider p-values significant.
        Will be marked with "*".
    test_lower_threshold: float
        Secondary theshold to consider p-values highly significant.
        Will be marked with "**".
    plot_non_significant: bool
        Whether to add a "n.s." sign to p-values above `test_upper_threshold`.
    plot_kws: dict
        Additional values to pass to seaborn.boxenplot or seaborn.swarmplot
    test_kws: dict
        Additional values to pass to pingouin.pairwise_tests.
        The default is: dict(parametric=False) to run a non-parametric test.

    Returns
    -------
    tuple[Figure, pandas.DataFrame]:
        if `ax` is None and `test` is True.

        pandas.DataFrame: if `ax` is not None.
        Figure: if `test` is False.
    None:
        if `test` is False and `ax` is not None.

    Raises
    ------
    ValueError:
        If either the `x` or `hue` column in `data` are not
        Category, string or object type, or if `y` is not numeric.

    """
    # opts = dict(data=data, x='h', y='y', hue='x', test_kws=dict(parametric=False))
    # opts = dict(data=data, x='cat', y='cont')
    # for k, v in opts.items():
    #     locals()[k] = v

    for var, name in [(x, "x"), (hue, "hue")]:
        if var is not None:
            if not data[var].dtype.name in ["category", "string", "object"]:
                raise ValueError(
                    f"`{name}` variable must be categorical, string or object."
                )

    if test_kws is None:
        test_kws = dict()
    if plot_kws is None:
        plot_kws = dict()

    data = data.sort_values([x] + ([hue] if hue is not None else []))

    if isinstance(y, (list, pd.Series, pd.Index)):
        if plot:
            # TODO: display only one legend for hue
            if ax is None:
                n, m = get_grid_dims(y)
                default_fig_kws = dict(
                    nrows=n, ncols=m, figsize=(m * 4, n * 4), sharex=True, squeeze=False
                )
                default_fig_kws.update(fig_kws or {})
                fig, axes = plt.subplots(**default_fig_kws)
                axes = axes.flatten()
            elif isinstance(ax, np.ndarray):
                axes = ax.flatten()
            elif isinstance(ax, matplotlib.axes.Axes):
                axes = np.asarray([ax])
        else:
            axes = [None] * len(y)

        _stats = list()
        idx = -1
        for idx, _var in enumerate(y):
            _ax = axes[idx]
            s: DataFrame = swarmboxenplot(
                data=data,
                x=x,
                y=_var,
                hue=hue,
                swarm=swarm,
                boxen=boxen,
                bar=bar,
                orient=orient,
                plot=plot,
                ax=_ax,
                test=test,
                to_test=to_test,
                multiple_testing=multiple_testing,
                test_upper_threshold=test_upper_threshold,
                test_lower_threshold=test_lower_threshold,
                plot_non_significant=plot_non_significant,
                plot_kws=plot_kws,
                test_kws=test_kws,
            )
            if plot:
                _ax.set(title=_var + _ax.get_title(), xlabel=None, ylabel=None)
            if test is not False:
                _stats.append(s.assign(Variable=_var))
        # "close" excess subplots
        if plot:
            for _ax in axes[idx + 1 :]:
                _ax.axis("off")
        if test is not False:
            stats = pd.concat(_stats).reset_index(drop=True)
            cols = [c for c in stats.columns if c != "Variable"]
            stats = stats.reindex(["Variable"] + cols, axis=1)

            # If there is just one test per `y` (no hue), correct p-values
            if stats.shape[0] == len(y):
                stats["p-cor"] = pg.multicomp(
                    stats["p-unc"].tolist(), method=multiple_testing
                )[1]
        if ax is None:
            return stats if not plot else (fig, stats) if test else fig
        return stats if test else None

    if data[y].dtype.name in ["category", "string", "object"]:
        raise ValueError("`y` variable must be numeric.")

    horizontal = orient in ["horizontal", "horiz", "h"]
    if horizontal:
        x2 = y
        y2 = x
        x = x2
        y = y2

    # Plot vanilla seaborn
    if plot:
        if ax is None:
            default_fig_kws = dict(figsize=(4, 4))
            default_fig_kws.update(fig_kws or {})
            fig, _ax = plt.subplots(**default_fig_kws)
        else:
            _ax = ax

        if boxen:
            assert not bar
            # Tmp fix for lack of support for Pandas Int64 in boxenplot:
            if data[y].dtype.name == "Int64":
                data[y] = data[y].astype(float)
            boxen_kws = filter_kwargs_by_callable(plot_kws, sns.boxenplot)
            sns.boxenplot(data=data, x=x, y=y, hue=hue, ax=_ax, **boxen_kws)
        if bar:
            assert not boxen
            bar_kws = filter_kwargs_by_callable(plot_kws, sns.barplot)
            sns.barplot(data=data, x=x, y=y, hue=hue, ax=_ax, **bar_kws)

        if (boxen or bar) and swarm:
            _add_transparency_to_plot(_ax, kind="bar" if bar else "boxen")
        if swarm:
            swarm_kws = filter_kwargs_by_callable(plot_kws, sns.swarmplot)
            if hue is not None and "dodge" not in swarm_kws:
                swarm_kws["dodge"] = True
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=UserWarning)
                sns.swarmplot(data=data, x=x, y=y, hue=hue, ax=_ax, **swarm_kws)
        if horizontal:
            _ax.set_yticklabels(_ax.get_yticklabels(), rotation=0, ha="right")
        else:
            _ax.set_xticklabels(_ax.get_xticklabels(), rotation=90, ha="right")

    if test is False:
        return fig if ax is None else None

    # Perform testing
    if horizontal:
        x2 = y
        y2 = x
        x = x2
        y = y2

    if test in [True, "t-test", "mann-whitney"]:
        test_function = pg.pairwise_tests
        if test == "mann-whitney":
            test_kws["parametric"] = False
    elif test in ["kruskal"]:
        test_function = pg.kruskal
        assert hue is None, "If test is 'kruskal', 'hue' must be None."
    else:
        raise ValueError(f"Test type '{test}' not recognized.")
    #
    if not data.index.is_unique:
        print("Warning: dataframe contains a duplicated index.")

    # # remove NaNs
    datat = data.dropna(subset=[x, y] + ([hue] if hue is not None else []))
    # # remove categories with only one element
    keep = datat.groupby(x).size()[datat.groupby(x).size() > 1].index
    datat = datat.loc[datat[x].isin(keep), :]
    if datat[x].dtype.name == "category":
        datat[x] = datat[x].cat.remove_unused_categories()
    if plot:
        ylim = _ax.get_ylim()  # save original axis boundaries for later
        ylength = abs(ylim[1]) + (abs(ylim[0]) if ylim[0] < 0 else 0)

    # # Now calculate stats
    # # # get empty dataframe in case nothing can be calculated
    stat = _get_empty_stat_results(datat, x, y, hue, add_median=True)

    # # # mirror groups to account for own pingouin order
    tats = stat.rename(
        columns={
            "B": "A",
            "A": "B",
            "median_A": "median_B",
            "median_B": "median_A",
        }
    )
    stat = (
        pd.concat([stat, tats])
        .sort_values(["Contrast", "A", "B"])
        .reset_index(drop=True)
    )
    try:
        _stat = test_function(
            data=datat,
            dv=y,
            between=x if hue is None else [x, hue],
            **test_kws,
        )
    except (AssertionError, ValueError) as e:
        print(str(e))
        _stat = stat
    except KeyError:
        print("Only one category with values!")
        _stat = stat

    if test == "kruskal":
        p = _stat.squeeze()["p-unc"]
        symbol = (
            "**"
            if p <= test_lower_threshold
            else "n.s."
            if ((p > test_upper_threshold) or pd.isnull(p))
            else "*"
        )
        _ax.set_title(symbol)
        return (fig, _stat) if ax is None else _stat

    stat = _stat.merge(
        stat[
            ["Contrast", "A", "B", "median_A", "median_B"]
            + ([x] if hue is not None else [])
        ],
        how="left",
    ).convert_dtypes()

    if to_test == "hue":
        stat = stat.loc[stat[x] != "-", :]

    if multiple_testing is not False:
        if "p-unc" not in stat.columns:
            stat["p-unc"] = np.nan
        stat["p-cor"] = pg.multicomp(
            stat["p-unc"].astype(float).values, method=multiple_testing
        )[1]
        pcol = "p-cor"
    else:
        pcol = "p-unc"

    if not plot:
        return stat

    # Plot
    # # This ensures there is a point for each `x` class and keep the order correct for below
    mm = data.groupby([x] + ([hue] if hue is not None else []))[y].median()
    if hue is None:
        order = {k: float(i) for i, k in enumerate(mm.index)}
    else:
        nhues = data[hue].drop_duplicates().dropna().shape[0]
        order = {
            k: (float(i) / nhues) - (1 / nhues) - 0.05 for i, k in enumerate(mm.index)
        }
    if horizontal:
        _ax.scatter(mm, order.values(), alpha=0, color="white")
    else:
        _ax.scatter(order.values(), mm, alpha=0, color="white")

    # # Plot significance bars
    # # # start at top of the plot and progressively decrease sig. bar downwards
    py = data[y].max()
    incr = ylength / 100  # divide yaxis in 100 steps
    for idx, row in stat.iterrows():
        p = row[pcol]
        if (pd.isnull(p) or (p > test_upper_threshold)) and (not plot_non_significant):
            py -= incr
            continue
        symbol = (
            "**"
            if p <= test_lower_threshold
            else "n.s."
            if ((p > test_upper_threshold) or pd.isnull(p))
            else "*"
        )
        if hue is not None:
            if row[x] != "-":
                xx = (order[(row[x], row["A"])], order[(row[x], row["B"])])
            else:
                try:
                    # TODO: get more accurate middle of group
                    xx = (
                        order[(row["A"], stat["A"].iloc[-1])] - (1 / nhues),
                        order[(row["B"], stat["B"].iloc[-1])] - (1 / nhues),
                    )
                except KeyError:
                    # These are the hue groups without contrasting on 'x'
                    continue
        else:
            xx = (order[row["A"]], order[row["B"]])

        _tp = (0.35 + xx[0], 0.35 + xx[1] - 0.25), (py, py)
        _tp2 = xx[1] - 0.025, py
        if horizontal:
            _tp = _tp[::-1]
            _tp2 = _tp2[::-1]

        _ax.plot(*_tp, color="black", linewidth=1.2)
        _ax.text(
            *_tp2,
            s=symbol,
            color="black",
            ha="center",
            rotation=90 if orient in ["horizontal", "horiz", "h"] else 0,
        )
        py -= incr
    _ax.set_ylim(ylim)
    return (fig, stat) if ax is None else stat


def _add_transparency_to_plot(
    ax: Axis, alpha: float = 0.25, kind: str = "boxen"
) -> None:

    objs = (
        (
            matplotlib.collections.PatchCollection,
            matplotlib.collections.PathCollection,
        )
        if kind == "boxen"
        else (matplotlib.patches.Rectangle)
    )

    for x in ax.get_children():
        if isinstance(x, objs):
            x.set_alpha(alpha)


def _get_empty_stat_results(
    data: DataFrame,
    x: str,
    y: str,
    hue: tp.Optional[str] = None,
    add_median: bool = True,
) -> DataFrame:
    stat = pd.DataFrame(
        itertools.combinations(data[x].drop_duplicates(), 2),
        columns=["A", "B"],
    )
    stat["Contrast"] = x
    if hue is not None:
        huestat = pd.DataFrame(
            itertools.combinations(data[hue].drop_duplicates(), 2),
            columns=["A", "B"],
        )
        huestat["Contrast"] = hue
        huestat[x] = "-"
        _to_append = [huestat]
        for v in data[x].unique():
            n = huestat.copy()
            n[x] = v
            n["Contrast"] = f"{x} * {hue}"
            _to_append.append(n)
        to_append = pd.concat(_to_append)
        stat = pd.concat([stat, to_append]).sort_values([x, "A", "B"])
        stat[x] = stat[x].fillna("-")
    stat["Tested"] = False
    stat["p-unc"] = np.nan

    if add_median:
        _mm = [data.groupby(x)[y].median().reset_index()]
        if hue is not None:
            _mm[0] = _mm[0].rename(columns={x: hue})
            _mm.append(data.groupby(hue)[y].median().reset_index())
            _p = data.groupby([x, hue])[y].median().reset_index()
            # remove categories if existing (workaround):
            _p = pd.DataFrame(_p.values, index=_p.index, columns=_p.columns)
            _mm.append(_p)
        mm = pd.concat(_mm)
        if mm[x].dtype.name == "category":
            mm[x] = mm[x].cat.add_categories(["-"]).fillna("-")
        else:
            mm[x] = mm[x].fillna("-")
        # mm = mm.append(data.groupby([x, hue])[y].std().reset_index()).fillna("-")
        for col in ["A", "B"]:
            stat = stat.merge(
                mm.rename(
                    columns={
                        hue if hue is not None else x: f"{col}",
                        y: f"median_{col}",
                    }
                ),
                how="left",
            )
    return stat



import typing as tp

import matplotlib.pyplot as plt

# from plottify import autosize

from seaborn_extensions.types import Figure, Axis, Series


def rankplot(
    series: Series,
    annotate_text: tp.Union[bool, tp.Sequence[str]] = True,
    n_top: int = 10,
    diff_threshold: tp.Optional[float] = None,
    fig_kws: tp.Dict = None,
    scatter_kws: tp.Dict = None,
    ax_kws: tp.Dict = None,
    ax: Axis = None,
) -> Figure:
    """ """
    if fig_kws is None:
        fig_kws = dict(figsize=(6, 4))
        fig_kws = dict()
    if scatter_kws is None:
        scatter_kws = dict()
    if ax_kws is None:
        ax_kws = dict(xlabel="Rank", ylabel="Value")
    if ax is None:
        fig, ax = plt.subplots(**fig_kws)
    else:
        fig = ax.figure
    rank = series.rank(method="dense")
    v = series.abs().max()
    ax.scatter(rank, series, c=series, cmap="coolwarm", vmin=-v, vmax=v, **scatter_kws)

    if annotate_text is not False:
        if annotate_text is True:
            if diff_threshold is not None:
                ts = series[series.abs() > diff_threshold].index.tolist()
            else:
                ts = series.sort_values().head(n_top // 2).index.tolist()
                ts += series.sort_values().tail(n_top // 2).index.tolist()
        else:
            ts = series.loc[annotate_text].index.tolist()
        for t in ts:
            ax.text(
                rank.loc[t],
                series.loc[t],
                s=t,
                ha="left" if series.loc[t] > 0 else "right",
            )
    ax.set(**ax_kws)
    # autosize(fig)
    ax.axhline(0, color="grey", linestyle="--", linewidth=1)
    return fig



import numpy as np
import matplotlib.pyplot as plt


def bubble_plot(corr):
    # Plot as miracle plot
    p = corr.reset_index().melt(id_vars="drug_pathway")

    fig, ax = plt.subplots(figsize=(8, 4))
    ax.scatter(
        p["expression_pathway"],
        p["drug_pathway"],
        c=p["value"],
        cmap="RdBu_r",
        s=5 + np.e ** 6 ** p["value"].abs(),
        alpha=0.5,
    )
    ax.set_xticklabels(p["expression_pathway"].unique(), rotation=90)
    return fig



"""
A replacement of seaborn.clustermap with additional features.
"""


import typing as tp
from typing import Optional, List, Union, Callable
import warnings

import numpy as np
import pandas as pd
import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns

from seaborn_extensions import SEQUENCIAL_CMAPS
from seaborn_extensions.types import Series, DataFrame, Array, Figure
from seaborn_extensions.utils import (
    minmax_scale,
    is_numeric,
    is_datetime,
    to_numeric,
    get_categorical_cmap,
    to_color_dataframe,
)


# TODO: revamp supporting custom cmaps/palettes


def clustermap(*args, **kwargs):
    # Defaults

    # # Size of figure
    if "figsize" not in kwargs:
        kwargs["figsize"] = (10, 10)
    else:
        if "square" in kwargs:
            if kwargs["square"] is True:
                print("`square` shape requested but `figsize` given. Ignoring `figsize`.")
    if kwargs["figsize"] == (10, 10):  # default value
        # assumes pivot_kws is not used...
        # would depend on x/y-ticklabel size...
        ...

    # # Decide if labeling x/y-ticklabels based on shape
    max_items = 120
    data = args[0]
    if "xticklabels" not in kwargs:
        kwargs["xticklabels"] = data.shape[1] < max_items
    if "yticklabels" not in kwargs:
        kwargs["yticklabels"] = data.shape[0] < max_items

    # dendrogram aspect ratio
    d = 0.1
    aspect = kwargs["figsize"][0] / kwargs["figsize"][1]
    smallest = (
        np.argmin(kwargs["figsize"]) if len(np.unique(kwargs["figsize"])) > 1 else -1
    )
    if smallest == -1:
        s = 1
        dar = (d, d)
    else:
        s = kwargs["figsize"][smallest] * d
        dar = tuple(d if i == smallest else s / kwargs["figsize"][i] for i in range(2))

    # # non-Z-score mode:
    nz_default_kws = dict(
        cmap="Reds",
        robust=True,
        dendrogram_ratio=dar,
        metric="correlation",
        square=True,
    )
    # # Z-score mode:
    zs_default_kws = dict(
        z_score=1,
        center=0,
        cmap="RdBu_r",
        robust=True,
        cbar_kws=dict(label="Z-score"),
        dendrogram_ratio=dar,
        metric="correlation",
        square=True,
    )
    if "config" in kwargs:
        default_kws = (
            zs_default_kws
            if kwargs["config"].lower() in ["z", "zscore", "z_score", "z-score"]
            else nz_default_kws
        )
        # kwargs.update(default_kws)  # for overwrite
        for k, v in default_kws.items():
            if k not in kwargs:
                kwargs[k] = v
        del kwargs["config"]
    if "cbar_kws" not in kwargs:
        kwargs["cbar_kws"] = dict()
    if smallest == 0:
        kwargs["cbar_kws"].update(dict(aspect=20 / aspect))

    # Square
    if "square" in kwargs:
        if kwargs["square"] is True:
            dw, dh = args[0].shape[::-1]
            dw *= 0.15
            dh *= 0.15
            th, tw = (
                args[0].index.to_series().astype(str).apply(len).max(),
                args[0].columns.to_series().astype(str).apply(len).max(),
            )
            tw *= 0.15
            th *= 0.15
            kwargs["figsize"] = (3 + dw + tw, 3 + dh + th)
        del kwargs["square"]

    # Annotations:
    cmaps = {"row": None, "col": None}
    # # capture "row_cmaps" and "col_cmaps" out of the kwargs
    for arg in ["row", "col"]:
        if arg + "_colors_cmaps" in kwargs:
            # TODO: make sure this matches in type/length the row/col_colors kwargs.
            cmaps[arg] = kwargs[arg + "_colors_cmaps"]
            del kwargs[arg + "_colors_cmaps"]

    # # get dataframe with colors and respective colormaps for rows and cols
    # # instead of the original numerical values
    _kwargs = dict(rows=None, cols=None)
    for arg in ["row", "col"]:
        if arg + "_colors" in kwargs:
            if isinstance(kwargs[arg + "_colors"], (pd.DataFrame, pd.Series)):
                _kwargs[arg + "s"] = kwargs[arg + "_colors"]
                kwargs[arg + "_colors"] = to_color_dataframe(
                    x=kwargs[arg + "_colors"],
                    cmaps=cmaps[arg],
                    offset=1 if arg == "row" else 0,
                )

    # Add p-value annotation
    # TODO: document
    if "pvalues" in kwargs:
        assert "annot" not in kwargs, "If providing p-values, `annot` cannot be used!"
        p = kwargs["pvalues"]
        # TODO: allow custom thresholds
        p = ((p < 0.05) & (p > 0.01)).replace({True: 1}) + ((p < 0.01)).replace({True: 2})
        kwargs["annot"] = p

    # Call original function
    grid = sns.clustermap(*args, **kwargs)

    # Add the colorbar legends to the figure
    _add_colorbars(grid, **_kwargs, row_cmaps=cmaps["row"], col_cmaps=cmaps["col"])

    # Some niceties
    ax = grid.ax_heatmap
    ax.set_xlabel(f"{ax.get_xlabel()}\n(n = {data.shape[1]})")
    ax.set_ylabel(f"{ax.get_ylabel()}\n(n = {data.shape[0]})")
    ax.set_yticklabels(ax.get_yticklabels(), rotation=0)
    ax.set_xticklabels(ax.get_xticklabels(), rotation=90)

    # Convert numeric p-value annotation to text
    if "pvalues" in kwargs:
        r = {"0": "", "1": "*", "2": "**"}
        for c in ax.get_children():
            if isinstance(c, matplotlib.text.Text):
                if c.get_text() in r:
                    c.set_text(r[c.get_text()])

    return grid


def _add_extra_colorbars_to_clustermap(
    grid: sns.matrix.ClusterGrid,
    datas: Union[Series, DataFrame],
    cmaps: Optional[Union[str, List[str]]] = None,
    # location: Union[Literal["col"], Literal["row"]] = "row",
    location: str = "row",
) -> None:
    """Add either a row or column colorbar to a seaborn Grid."""

    def add(data: Series, cmap: str, bbox: List[List[int]], orientation: str) -> None:
        ax = grid.fig.add_axes(matplotlib.transforms.Bbox(bbox))
        if is_numeric(data):
            if is_datetime(data):
                data = minmax_scale(data)
            norm = matplotlib.colors.Normalize(vmin=data.min(), vmax=data.max())
            cbar = matplotlib.colorbar.ColorbarBase(
                ax,
                cmap=plt.get_cmap(cmap),
                norm=norm,
                orientation=orientation,
                label=data.name,
            )
        else:
            res = to_numeric(data)
            # res /= res.max()
            cmap = get_categorical_cmap(res)
            # norm = matplotlib.colors.Normalize(vmin=res.min(), vmax=res.max())
            cbar = matplotlib.colorbar.ColorbarBase(
                ax,
                cmap=cmap,
                orientation=orientation,
                label=data.name,
            )
            cbar.set_ticks(res.drop_duplicates().sort_values() / res.max())
            if orientation == "vertical":
                cbar.ax.set_xticklabels(
                    data.value_counts().sort_index().index, rotation=0
                )
            else:
                cbar.ax.set_xticklabels(
                    data.value_counts().sort_index().index, rotation=90
                )

    offset = 1 if location == "row" else 0

    if isinstance(datas, pd.Series):
        datas = datas.to_frame()
    if cmaps is None:
        cmaps = SEQUENCIAL_CMAPS[offset:]
    if isinstance(cmaps, str):
        cmaps = [cmaps]

    # get position to add new axis in existing figure
    # # get_position() returns ((x0, y0), (x1, y1))
    heat = grid.ax_heatmap.get_position()
    cbar_spacing = 0.05
    cbar_size = 0.025
    if location == "col":
        orientation = "vertical"
        dend = grid.ax_col_dendrogram.get_position()
        y0 = dend.y0
        y1 = dend.y1
        for i, (data, cmap) in enumerate(zip(datas, cmaps)):
            if i == 0:
                x0 = heat.x1
                x1 = heat.x1 + cbar_size
            else:
                x0 += cbar_size + cbar_spacing
                x1 += cbar_size + cbar_spacing
            add(datas[data], cmap, [[x0, y0], [x1, y1]], orientation)
    else:
        orientation = "horizontal"
        dend = grid.ax_row_dendrogram.get_position()
        x0 = dend.x0
        x1 = dend.x1
        for i, (data, cmap) in enumerate(zip(datas, cmaps)):
            if i == 0:
                y0 = dend.y0 - cbar_size
                y1 = dend.y0
            else:
                y0 -= cbar_size + cbar_spacing
                y1 -= cbar_size + cbar_spacing
            add(datas[data], cmap, [[x0, y0], [x1, y1]], orientation)


def _add_colorbars(
    grid: sns.matrix.ClusterGrid,
    rows: DataFrame = None,
    cols: DataFrame = None,
    row_cmaps: Optional[List[str]] = None,
    col_cmaps: Optional[List[str]] = None,
) -> None:
    """Add row and column colorbars to a seaborn Grid."""
    if rows is not None:
        _add_extra_colorbars_to_clustermap(grid, rows, location="row", cmaps=row_cmaps)
    if cols is not None:
        _add_extra_colorbars_to_clustermap(grid, cols, location="col", cmaps=col_cmaps)


def _add_docs_to_clustermap():
    """
    Edit original seaborn.clustermap docstring to document {row,col}_colors_cmaps arguments.
    """
    # TODO: finish documenting changes.
    error_msg = (
        "Seaborn version may not be compatible with seaborn_extensions version."
        "Skipping annotating clustermap function docstring."
    )

    docs = sns.clustermap.__doc__
    anchors = np.asarray(
        [
            ("pivot_kws : ", "method : "),
            ("{row,col}_colors : ", "mask : bool"),
            ("kwargs : other keyword arguments", "Returns"),
        ]
    )

    points = np.zeros(anchors.shape, dtype=int)
    for i, tup in enumerate(anchors):
        for j, p in enumerate(tup):
            try:
                x = docs.index(p)
            except ValueError:
                print(error_msg)
                return
            points[i, j] = x

    add_docs1 = """config : str, optional
        EXTENSION!
        One of two pre-defined configurations: "abs", "zscore".
        These two configurations provide custom default keyword arguments
        compared with the native seaborn function and several adjustments to
        figure and axis sizes, labels and other objects.
        Options:
         - "abs": good for non-negative data.
         - "zscore": good for real data with variables with very different means.
        Other keyword arguments affected (only is not provided):
         - {x,y}ticklabels: will turn off if more than 120 items in each axis.
         - dendrogram_ratio: will adjust, given relative shape of data.
    """
    add_docs2 = """{row,col}_colors : list-like or pandas DataFrame/Series, optional
        EXTENSION!
        List of colors to label for either the rows or columns. Useful to
        evaluate whether samples within a group are clustered together. Can
        use nested lists or DataFrame for multiple color levels of labeling.
        If given as a DataFrame or Series, labels for the colors are extracted
        from the DataFrames column names or from the name of the Series.
        DataFrame/Series colors are also matched to the data by their
        index, ensuring colors are drawn in the correct order.

        TODO: complete defining new behavious
    {row,col}_colors_cmaps: Sequence[str]
        EXTENSION!
        Colormaps to be used for the variables provided in `{row,col}_colors`.
    """
    add_docs3 = """pvalues : pandas DataFrame, optional
        EXTENSION!
        A dataframe matching the input shape, where the values are p-values.
        Values 0.05 > p > 0.01 will be labeled with '*'.
        Values p < 0.01 will be labeled with '**'.
        Values p >= 0.05 will not be labeled.
        This will be overlaid as text on top of the heatmap.
        If providing `pvalues`, `annot` cannot be used.
    square: bool, optional
        EXTENSION!
        Try to make the shape of the figure as square as possible.
        If used, `figsize` will be ignored.

    """
    clustermap.__doc__ = (
        docs[: points[0][0]]
        + add_docs1
        + docs[points[0][1] : points[1][0]]
        + add_docs2
        + docs[points[1][1] : points[2][0]]
        + add_docs3
        + docs[points[2][1] :]
    )


_add_docs_to_clustermap()


def colorbar_decorator(f: Callable) -> Callable:
    """
    Decorate seaborn.clustermap in order to have numeric values passed to the
    ``row_colors`` and ``col_colors`` arguments translated into row and column
    annotations and in addition colorbars for the restpective values.
    """
    # Add a flag
    f.decorated = True

    return clustermap


def activate():
    warnings.warn(
        "Decoration of native searborn.clustermap function will be deprecated in version 1.0.0, use 'from seaborn_extensions import clustermap' instead.",
        PendingDeprecationWarning,
    )
    if sns.clustermap.__module__ != "seaborn_extensions.annotated_clustermap":
        sns.clustermap = colorbar_decorator(sns.clustermap)


# To plot just the attribute heatmap:
def get_attribute_colors(
    y: DataFrame,
    attributes: tp.Sequence[str],
    palettes: tp.Mapping[str, tp.Tuple[float]],
    cmaps: tp.Mapping[str, str],
    as_dataframe: bool = False,
) -> tp.Union[Array, DataFrame]:
    vals = list()
    for attr in attributes:
        if attr in palettes:
            p = dict(zip(y[attr].cat.categories, palettes[attr]))
            val = np.asarray([p[v] if not pd.isnull(v) else (0, 0, 0) for v in y[attr]])
        elif attr in cmaps:
            cmap = plt.get_cmap(cmaps[attr])
            val = cmap(minmax_scale(y[attr].astype(float)))[:, :3]
        vals.append(val)
    if as_dataframe:
        return pd.DataFrame(
            map(tuple, np.asarray(vals)), index=attributes, columns=y.index
        )
    return np.asarray(vals)


def plot_attribute_heatmap(
    y: DataFrame,
    attributes: tp.Sequence[str],
    palettes: tp.Mapping[str, tp.Tuple[float]],
    cmaps: tp.Mapping[str, str],
    **kwargs,
) -> Figure:
    vals = get_attribute_colors(y, attributes, palettes, cmaps)
    if "ax" not in kwargs:
        fig, axes = plt.subplots(
            len(attributes), **kwargs, gridspec_kw=dict(wspace=0, hspace=0)
        )
    else:
        fig = kwargs["ax"].figure
    # ax.imshow(vals)
    for _p, attr, ax in zip(vals, attributes, axes):
        ax.imshow(_p[np.newaxis, ...])
        ax.set(xticks=[], yticks=[0])
        ax.set_yticklabels([attr], rotation=0)
        sns.despine(ax=ax, left=True, bottom=True)
    ax = axes[-1]
    ax.set_xticks(range(len(y.index)))
    ax.set_xticklabels(y.index, rotation=90)
    return fig



import typing as tp
from collections.abc import MutableSequence

import pandas
import numpy
import matplotlib

# type aliasing (done with Union to distinguish from other declared variables)
Array = numpy.ndarray
Series = pandas.Series
MultiIndexSeries = pandas.Series
DataFrame = pandas.DataFrame

Figure = matplotlib.figure.Figure
Axis = matplotlib.axis.Axis
Patch = matplotlib.patches.Patch
ColorMap = matplotlib.colors.LinearSegmentedColormap

Iterables = tp.Union[set, MutableSequence, pandas.Series, pandas.Index]



"""Top-level package for seaborn_extensions."""

from outdated import warn_if_outdated

try:
    # Even though there is no "../_version" file,
    # it should be generated by
    # setuptools_scm when building the package
    from seaborn_extensions._version import __version__

    version = __version__
except ImportError:
    from setuptools_scm import get_version as _get_version

    version = __version__ = _get_version(root="..", relative_to=__file__)

__author__ = """Andre Rendeiro"""
__email__ = "afrendeiro@gmail.com"


warn_if_outdated("seaborn_extensions", __version__)


DEFAULT_CHANNEL_COLORS = [
    "red",
    "green",
    "blue",
    "orange",
    "purple",
    "brown",
    "pink",
    "olive",
    "cyan",
    "gray",
]

SEQUENCIAL_CMAPS = [
    "Purples",
    "Greens",
    "Oranges",
    "Greys",
    "Reds",
    "Blues",
    "YlOrBr",
    "YlOrRd",
    "OrRd",
    "PuRd",
    "RdPu",
    "BuPu",
    "GnBu",
    "PuBu",
    "YlGnBu",
    "PuBuGn",
    "BuGn",
    "YlGn",
    "binary",
    "gist_yarg",
    "gist_gray",
    "gray",
    "bone",
    "pink",
    "spring",
    "summer",
    "autumn",
    "winter",
    "cool",
    "Wistia",
    "hot",
    "afmhot",
    "gist_heat",
    "copper",
]


from seaborn_extensions.swarmboxenplot import swarmboxenplot
from seaborn_extensions.annotated_clustermap import (
    activate as activate_annotated_clustermap,
    clustermap,
)
from seaborn_extensions.volcano import volcano_plot
from seaborn_extensions.rankplot import rankplot



import typing as tp

import numpy as np
import matplotlib.pyplot as plt
import pingouin as pg

from seaborn_extensions.types import Figure, Axis, DataFrame
from seaborn_extensions.utils import get_grid_dims


def volcano_plot(
    stats: DataFrame,
    annotate_text: tp.Union[bool, tp.Sequence[str]] = True,
    diff_threshold: tp.Optional[float] = 0.05,
    n_top: int = None,
    invert_direction: bool = True,
    fig_kws: tp.Dict = None,
    axes: tp.Sequence[Axis] = None,
) -> Figure:
    """
    Assumes stats dataframe from seaborn_extensions.swarmboxenplot:
        - "hedges/coefs" column with effect sizes
        - "p-unc/pvalues" column with significance
        - "p-cor" column with significance corrected for multiple testing (will be added if missing)
        - "Variable" column with variable names (will use dataframe index if missing)
        If multiple tests are performed, each will be plotted in a subplot:
        - "A", "B" group identifiers such that hedges is positive value ~if mean(A) > mean(B).
    """
    if diff_threshold is not None:
        assert n_top is None
    else:
        assert n_top is not None

    if "hedges" not in stats.columns and "coefs" in stats.columns:
        print("Using columns 'coefs' as effect size estimates.")
        stats["hedges"] = stats["coefs"]
    if "p-unc" not in stats.columns and "pvalues" in stats.columns:
        print("Using columns 'pvalues' as effect size estimates.")
        stats["p-unc"] = stats["pvalues"]
    if "p-cor" not in stats.columns:
        print("Corrected p-values not given, using BH-FDR method.")
        stats["p-cor"] = pg.multicomp(stats["p-unc"], method="fdr_bh")[1]

    if (stats[["p-unc", "p-cor"]] == 0).any().any():
        print("p-values include zeros, replacing values around 1e-300 for display.")
        sel = stats["p-unc"] == 0
        stats.loc[sel, "p-unc"] = np.exp(-np.random.randint(280, 300, sel.sum())) ** 2.3
        stats.loc[stats["p-cor"] == 0, "p-cor"] = (
            np.exp(-np.random.randint(280, 300, sel.sum())) ** 2.3
        )

    if "Variable" not in stats.columns:
        print("Using dataframe index as variable names.")
        stats["Variable"] = stats.index

    for col in ["A", "B"]:
        if col not in stats.columns:
            stats[col] = col

    stats["A"] = stats["A"].astype(str)
    stats["B"] = stats["B"].astype(str)
    combs = stats[["A", "B"]].drop_duplicates().reset_index(drop=True)
    if invert_direction:
        stats["hedges"] *= -1  # convert to B / A which is often more intuitive
    stats["logp-unc"] = -np.log10(stats["p-unc"].fillna(1))
    stats["logp-cor"] = -np.log10(stats["p-cor"].fillna(1))
    stats["p-cor-plot"] = (stats["logp-cor"] / stats["logp-cor"].max()).fillna(1) * 5
    n, m = get_grid_dims(combs.shape[0])
    default_kws = dict(nrows=n, ncols=m, figsize=(4 * m, 4 * n), squeeze=False)
    default_kws.update(fig_kws or {})
    if axes is None:
        fig, axes = plt.subplots(**default_kws)
    else:
        axes = np.asarray(axes)
        if len(axes.shape) == 1:
            axes = axes.reshape((-1, 1))
        fig = axes.flatten()[0].figure
    idx = -1
    for idx, (a, b) in combs.iterrows():
        ax = axes.flatten()[idx]
        p = stats.query(f"A == '{a}' & B == '{b}'")
        ax.axvline(0, linestyle="--", color="grey")
        v = p["hedges"].abs().max()
        ax.scatter(
            p["hedges"],
            p["logp-unc"],
            c=p["hedges"],
            s=5 + (2 ** p["p-cor-plot"]),
            cmap="coolwarm",
            vmin=-v,
            vmax=v,
            rasterized=True,
        )
        ax.set(title=f"{b} / {a}", ylabel=None, xlabel=None)
        if annotate_text is not False:
            if annotate_text is True:
                if diff_threshold is not None:
                    ts = p.query(f"`p-cor` < {diff_threshold}").index
                else:
                    ts = p.sort_values("p-unc").head(n_top).index
            else:
                ts = p.loc[p["Variable"].isin(annotate_text)].index
            for t in ts:
                ax.text(
                    p.loc[t, "hedges"],
                    p.loc[t, "logp-unc"],
                    s=p.loc[t, "Variable"],
                    ha="left" if p.loc[t, "hedges"] > 0 else "right",
                )
    for ax in axes.flatten()[idx + 1 :]:
        ax.axis("off")

    for ax in axes[:, 0]:
        ax.set(ylabel="-log10(p-val)")
    for ax in axes[-1, :]:
        ax.set(xlabel="Hedges' g")

    if "p-cor-plot" in stats.columns:
        stats = stats.drop("p-cor-plot", axis=1)
    return fig
